"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.transformStyles = exports.parseString = void 0;
const create_1 = __importDefault(require("twrnc/create"));
const crypto_1 = require("crypto");
/**
 * Matches balanced parentheses and content, with up to 5 levels of nesting.
 * Based on: https://stackoverflow.com/a/17759264
 */
const BALANCED_PARENTHESES_REGEX = /(?<prefix>[\w\d-]+(?::[\w\d-]+)?):\((?<content>([^()]*|\((?:[^()]*|\((?:[^()]*|\((?:[^()]*|\([^()]*\))*\))*\))*\))*)\)/g;
/**
 * Matches simple tailwind styles, like "pt-8" or "ios:pt-2"
 */
const REGEX = /(?:(?<prefix>(?:[\w\d-[\]]+:?)*):)?(?:(?<content>[\w\d-[\]]+))/g;
const PLATFORMS = new Set([
    "ios",
    "android",
    "web",
    "macos",
    "windows",
]);
const SELECTORS = new Set([
    "active",
    "hover",
    "focus",
    "focus-within",
    "focus-visible",
    "visited",
]);
/**
 * Parses a string recursively into an array of content + prefix combinations.
 *
 * @param str
 * @param parentPrefixes
 */
const parseString = (str, parentPrefixes = []) => {
    const nested = [...str.matchAll(BALANCED_PARENTHESES_REGEX)]
        .map((match) => ({
        prefixes: [
            ...parentPrefixes,
            ...match.groups.prefix.split(":").filter((prefix) => prefix.length),
        ],
        content: match.groups.content,
    }))
        .filter(({ content }) => content.length)
        // Recusively resolve parentheses content
        .flatMap(({ prefixes, content }) => (0, exports.parseString)(content, prefixes));
    // Remove nested styles
    const cleaned = str.replace(BALANCED_PARENTHESES_REGEX, "");
    const nonNested = [...cleaned.matchAll(REGEX)].map((match) => {
        var _a, _b;
        return ({
            prefixes: [
                ...parentPrefixes,
                ...((_b = (_a = match.groups.prefix) === null || _a === void 0 ? void 0 : _a.split(":").filter((prefix) => prefix.length)) !== null && _b !== void 0 ? _b : []),
            ],
            content: match.groups.content,
        });
    });
    return [...nonNested, ...nested];
};
exports.parseString = parseString;
/**
 * Groups all styles for the same prefix combinations.
 *
 * @param parsed
 */
const groupByPrefixes = (parsed) => {
    const grouped = parsed.reduce((acc, cur) => {
        const key = [...cur.prefixes].sort().join(":");
        if (!acc[key])
            acc[key] = [cur];
        else
            acc[key].push(cur);
        return acc;
    }, {});
    return Object.values(grouped).map((arr) => ({
        prefixes: arr[0].prefixes,
        styles: arr.map(({ content }) => content),
    }));
};
const resolvePrefixes = (prefixes, twConfig) => {
    var _a;
    // Breakpoints are sorted ascending in min-width
    const breakpoints = Object.keys((_a = twConfig.theme.screens) !== null && _a !== void 0 ? _a : {});
    // Find the largest specified breakpoint
    const breakpointIndex = prefixes.reduce((acc, cur) => {
        // Skip smaller breakpoints
        const index = breakpoints.indexOf(cur, acc > -1 ? acc : 0);
        if (index > acc)
            return index;
        else
            return acc;
    }, -1);
    const breakpoint = breakpointIndex > -1 ? breakpoints[breakpointIndex] : undefined;
    return {
        dark: prefixes.some((prefix) => prefix === "dark"),
        platform: prefixes.find((prefix) => PLATFORMS.has(prefix)),
        selectors: prefixes.filter((prefix) => SELECTORS.has(prefix)),
        breakpoint: breakpoint
            ? {
                label: breakpoint,
                // @ts-ignore
                minWidth: twConfig.theme.screens[breakpoint],
            }
            : undefined,
    };
};
/**
 * Translates a string of tailwind utility classes into an array of styles with resolved prefixes.
 *
 * @param str
 * @param twConfig
 */
const transformStyles = (str, twConfig) => {
    const parsed = (0, exports.parseString)(str);
    const grouped = groupByPrefixes(parsed);
    /**
     * Apply an arbitrary "ios" since we don't pass in styles that are prefixed.
     */
    const tw = (0, create_1.default)(twConfig, "ios");
    return grouped.map(({ prefixes, styles }) => {
        const resolvedPrefixes = resolvePrefixes(prefixes, twConfig);
        const resolvedStyle = styles
            .map((style) => tw `${style}`)
            .reduce((acc, cur) => Object.assign(acc, cur), {});
        return Object.assign(Object.assign({}, resolvedPrefixes), { id: `rntwm-${hash(JSON.stringify({ prefixes, resolvedStyle }))}`, style: resolvedStyle });
    });
};
exports.transformStyles = transformStyles;
const hash = (str) => (0, crypto_1.createHash)("md5")
    .update(str)
    .digest("base64")
    .replace(/\//g, "_")
    .replace(/[=]/g, "");
